import logging

from ..order.orderpackage import OrderPackageType, BaseOrder
from ..exceptions import ControlError

logger = logging.getLogger(__name__)


class BaseControl:

    NAME = None

    def __init__(self, flumine, *args, **kwargs):
        self.flumine = flumine

    def __call__(self, order: BaseOrder, package_type: OrderPackageType):
        self._validate(order, package_type)

    def _validate(self, order: BaseOrder, package_type: OrderPackageType) -> None:
        raise NotImplementedError

    def _on_error(self, order: BaseOrder, error: str) -> None:
        violation_msg = "Order has violated: {0} Error: {1}".format(self.NAME, error)
        order.violation(violation_msg)
        if logger.isEnabledFor(logging.WARNING):
            logger.warning(
                violation_msg,
                extra={"control": self.NAME, "error": error, "order": order.info},
            )
        raise ControlError(violation_msg)



import logging
import datetime
from typing import Optional
from betfairlightweight.streaming import StreamListener, HistoricalGeneratorStream
from betfairlightweight.streaming.stream import MarketStream, RaceStream
from betfairlightweight.streaming.cache import MarketBookCache, RaceCache
from betfairlightweight.resources.baseresource import BaseResource
from betfairlightweight.compat import json

from .basestream import BaseStream
from ..exceptions import ListenerError
from ..utils import create_time

logger = logging.getLogger(__name__)


class FlumineMarketStream(MarketStream):
    """
    Custom bflw stream to speed up processing
    by limiting to inplay/not inplay or limited
    seconds to start.
    `_process` updated to not call `on_process`
    which reduces some function calls.
    """

    def _process(self, data: list, publish_time: int) -> bool:
        active = False
        for market_book in data:
            if "id" not in market_book:
                continue
            market_id = market_book["id"]
            full_image = market_book.get("img", False)
            market_book_cache = self._caches.get(market_id)

            if (
                full_image or market_book_cache is None
            ):  # historic data does not contain img
                if "marketDefinition" not in market_book:
                    logger.warning(
                        "[%s: %s]: Missing marketDefinition on market %s resulting "
                        "in potential missing data in the MarketBook (make sure "
                        "EX_MARKET_DEF is requested)"
                        % (self, self.unique_id, market_id)
                    )
                market_book_cache = MarketBookCache(
                    market_id,
                    publish_time,
                    self._lightweight,
                    self._calculate_market_tv,
                    self._cumulative_runner_tv,
                )
                self._caches[market_id] = market_book_cache
                logger.info(
                    "[%s: %s]: %s added, %s markets in cache"
                    % (self, self.unique_id, market_id, len(self._caches))
                )

            # listener_kwargs filtering
            active = True
            if "marketDefinition" in market_book:
                _definition_status = market_book["marketDefinition"].get("status")
                _definition_in_play = market_book["marketDefinition"].get("inPlay")
                _definition_market_time = market_book["marketDefinition"].get(
                    "marketTime"
                )
            else:
                _definition_status = market_book_cache._definition_status
                _definition_in_play = market_book_cache._definition_in_play
                _definition_market_time = market_book_cache.market_definition[
                    "marketTime"
                ]

            # if market is not open (closed/suspended) process regardless
            if _definition_status == "OPEN":
                if self._listener.inplay:
                    if not _definition_in_play:
                        active = False
                elif self._listener.seconds_to_start:
                    _now = datetime.datetime.utcfromtimestamp(publish_time / 1e3)
                    _market_time = BaseResource.strip_datetime(_definition_market_time)
                    seconds_to_start = (_market_time - _now).total_seconds()
                    if seconds_to_start > self._listener.seconds_to_start:
                        active = False
                if self._listener.inplay is False:
                    if _definition_in_play:
                        active = False
            # check if refresh required
            if active and not market_book_cache.active:
                market_book_cache.refresh_cache()

            market_book_cache.update_cache(market_book, publish_time, active=active)
            self._updates_processed += 1
        return active


class FlumineRaceStream(RaceStream):
    """
    `_process` updated to not call `on_process`
    which reduces some function calls.
    """

    def _process(self, race_updates: list, publish_time: int) -> bool:
        active = False
        for update in race_updates:
            market_id = update["mid"]
            race_cache = self._caches.get(market_id)
            if race_cache is None:
                race_id = update.get("id")
                race_cache = RaceCache(
                    market_id, publish_time, race_id, self._lightweight
                )
                race_cache.start_time = create_time(publish_time, race_id)
                self._caches[market_id] = race_cache
                logger.info(
                    "[%s: %s]: %s added, %s markets in cache"
                    % (self, self.unique_id, market_id, len(self._caches))
                )

            # filter after start time
            diff = (
                race_cache.start_time
                - datetime.datetime.utcfromtimestamp(publish_time / 1e3)
            ).total_seconds()
            if diff <= 0:
                race_cache.update_cache(update, publish_time)
                self._updates_processed += 1
                active = True
        return active


class HistoricListener(StreamListener):
    """
    Custom listener to restrict processing by
    inplay or seconds_to_start.
    """

    def __init__(self, inplay: bool = None, seconds_to_start: float = None, **kwargs):
        super(HistoricListener, self).__init__(**kwargs)
        self.inplay = inplay
        self.seconds_to_start = seconds_to_start

    def _add_stream(self, unique_id: int, operation: str):
        if operation == "marketSubscription":
            return FlumineMarketStream(self, unique_id)
        elif operation == "orderSubscription":
            raise ListenerError("Unable to process order stream")
        elif operation == "raceSubscription":
            return FlumineRaceStream(self, unique_id)

    def on_data(self, raw_data: str) -> Optional[bool]:
        try:
            data = json.loads(raw_data)
        except ValueError:
            logger.error("value error: %s" % raw_data)
            return

        # remove error handler / operation check

        # skip on_change / on_update as we know it is always an update
        publish_time = data["pt"]
        return self.stream._process(data[self.stream._lookup], publish_time)


class FlumineHistoricalGeneratorStream(HistoricalGeneratorStream):
    """Super fast historical stream"""

    def _read_loop(self) -> dict:
        self.listener.register_stream(self.unique_id, self.operation)
        listener_on_data = self.listener.on_data  # cache functions
        stream_snap = self.listener.stream.snap
        with open(self.file_path, "r") as f:
            for update in f:
                if listener_on_data(update):
                    yield stream_snap()


class HistoricalStream(BaseStream):

    LISTENER = HistoricListener
    MAX_LATENCY = None

    def run(self) -> None:
        pass

    def handle_output(self) -> None:
        pass

    def create_generator(self):
        self._listener.update_clk = (
            False  # do not update clk on updates (not required when backtesting)
        )
        stream = FlumineHistoricalGeneratorStream(
            file_path=self.market_filter,
            listener=self._listener,
            operation=self.operation,
            unique_id=self.stream_id,
        )
        return stream.get_generator()



import unittest
from unittest import mock
from unittest.mock import call

from flumine.execution.transaction import Transaction, OrderPackageType
from flumine.exceptions import ControlError, OrderError


class TransactionTest(unittest.TestCase):
    def setUp(self) -> None:
        mock_blotter = {}
        self.mock_market = mock.Mock(blotter=mock_blotter)
        self.transaction = Transaction(self.mock_market, 1, False)

    def test_init(self):
        self.assertEqual(self.transaction.market, self.mock_market)
        self.assertEqual(self.transaction._id, 1)
        self.assertFalse(self.transaction._async_place_orders)
        self.assertFalse(self.transaction._pending_orders)
        self.assertEqual(self.transaction._pending_place, [])
        self.assertEqual(self.transaction._pending_cancel, [])
        self.assertEqual(self.transaction._pending_update, [])
        self.assertEqual(self.transaction._pending_replace, [])

    @mock.patch("flumine.execution.transaction.get_market_notes")
    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=True,
    )
    @mock.patch("flumine.execution.transaction.events")
    def test_place_order(
        self, mock_events, mock__validate_controls, mock_get_market_notes
    ):
        self.transaction.market.blotter = mock.MagicMock()
        self.transaction.market.blotter.has_trade.return_value = False
        mock_order = mock.Mock(id="123", lookup=(1, 2, 3))
        mock_order.trade.market_notes = None
        self.assertTrue(self.transaction.place_order(mock_order))
        mock_order.place.assert_called_with(
            self.transaction.market.market_book.publish_time, None, False
        )
        self.transaction.market.flumine.log_control.assert_called_with(
            mock_events.TradeEvent()
        )
        mock_get_market_notes.assert_called_with(
            self.mock_market, mock_order.selection_id
        )
        mock__validate_controls.assert_called_with(mock_order, OrderPackageType.PLACE)
        mock__validate_controls.assert_called_with(mock_order, OrderPackageType.PLACE)
        self.transaction._pending_place = [(mock_order, None)]
        self.assertTrue(self.transaction._pending_orders)
        mock_order.trade.strategy.get_runner_context.assert_called_with(
            *mock_order.lookup
        )
        self.transaction.market.blotter.has_trade.assert_called_with(
            mock_order.trade.id
        )

    @mock.patch("flumine.execution.transaction.get_market_notes")
    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=True,
    )
    @mock.patch("flumine.execution.transaction.events")
    def test_place_order_not_executed(
        self, mock_events, mock__validate_controls, mock_get_market_notes
    ):
        self.transaction.market.blotter = mock.MagicMock()
        self.transaction.market.blotter.has_trade.return_value = False
        mock_order = mock.Mock(id="123")
        self.assertTrue(self.transaction.place_order(mock_order, execute=False))
        mock_order.place.assert_called_with(
            self.transaction.market.market_book.publish_time, None, False
        )
        self.transaction.market.flumine.log_control.assert_called_with(
            mock_events.TradeEvent()
        )
        mock__validate_controls.assert_not_called()
        self.transaction._pending_place = []
        self.assertFalse(self.transaction._pending_orders)
        self.transaction.market.blotter.has_trade.assert_called_with(
            mock_order.trade.id
        )

    @mock.patch("flumine.execution.transaction.get_market_notes")
    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=True,
    )
    def test_place_order_retry(self, mock__validate_controls, mock_get_market_notes):
        self.transaction.market.blotter = mock.MagicMock()
        self.transaction.market.blotter.has_trade.return_value = False
        self.transaction.market.blotter.__contains__.return_value = True
        mock_order = mock.Mock(lookup=(1, 2, 3))
        with self.assertRaises(OrderError):
            self.transaction.place_order(mock_order)
        mock__validate_controls.assert_called_with(mock_order, OrderPackageType.PLACE)
        self.transaction._pending_place = []
        self.assertFalse(self.transaction._pending_orders)
        mock_get_market_notes.assert_not_called()

    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=False,
    )
    def test_place_order_violation(self, mock__validate_controls):
        mock_order = mock.Mock()
        self.assertFalse(self.transaction.place_order(mock_order))
        mock__validate_controls.assert_called_with(mock_order, OrderPackageType.PLACE)
        self.transaction._pending_place = []
        self.assertFalse(self.transaction._pending_orders)

    @mock.patch("flumine.execution.transaction.get_market_notes")
    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=False,
    )
    @mock.patch("flumine.execution.transaction.events")
    def test_force_place_order(
        self, mock_events, mock__validate_controls, mock_get_market_notes
    ):
        self.transaction.market.blotter = mock.MagicMock()
        self.transaction.market.blotter.has_trade.return_value = False
        mock_order = mock.Mock(id="123", lookup=(1, 2, 3))
        self.assertTrue(self.transaction.place_order(mock_order, force=True))
        mock__validate_controls.assert_not_called()
        self.transaction._pending_place = [(mock_order, None)]
        self.assertTrue(self.transaction._pending_orders)

    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=True,
    )
    def test_cancel_order(self, mock__validate_controls):
        mock_order = mock.Mock()
        self.assertTrue(self.transaction.cancel_order(mock_order, 0.01))
        mock_order.cancel.assert_called_with(0.01)
        mock__validate_controls.assert_called_with(mock_order, OrderPackageType.CANCEL)
        self.transaction._pending_cancel = [(mock_order, None)]
        self.assertTrue(self.transaction._pending_orders)

    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=False,
    )
    def test_cancel_order_violation(self, mock__validate_controls):
        mock_order = mock.Mock()
        self.assertFalse(self.transaction.cancel_order(mock_order))
        mock__validate_controls.assert_called_with(mock_order, OrderPackageType.CANCEL)
        self.transaction._pending_cancel = []
        self.assertFalse(self.transaction._pending_orders)

    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=False,
    )
    def test_force_cancel_order(self, mock__validate_controls):
        mock_order = mock.Mock()
        self.assertTrue(self.transaction.cancel_order(mock_order, 0.01, force=True))
        mock_order.cancel.assert_called_with(0.01)
        mock__validate_controls.assert_not_called()
        self.transaction._pending_cancel = [(mock_order, None)]
        self.assertTrue(self.transaction._pending_orders)

    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=True,
    )
    def test_update_order(self, mock__validate_controls):
        mock_order = mock.Mock()
        self.assertTrue(self.transaction.update_order(mock_order, "PERSIST"))
        mock_order.update.assert_called_with("PERSIST")
        mock__validate_controls.assert_called_with(mock_order, OrderPackageType.UPDATE)
        self.transaction._pending_update = [(mock_order, None)]
        self.assertTrue(self.transaction._pending_orders)

    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=False,
    )
    def test_update_order_violation(self, mock__validate_controls):
        mock_order = mock.Mock()
        self.assertFalse(self.transaction.update_order(mock_order, "test"))
        mock__validate_controls.assert_called_with(mock_order, OrderPackageType.UPDATE)
        self.transaction._pending_update = []
        self.assertFalse(self.transaction._pending_orders)

    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=False,
    )
    def test_force_update_order(self, mock__validate_controls):
        mock_order = mock.Mock()
        self.assertTrue(
            self.transaction.update_order(mock_order, "PERSIST", force=True)
        )
        mock_order.update.assert_called_with("PERSIST")
        mock__validate_controls.assert_not_called()
        self.transaction._pending_update = [(mock_order, None)]
        self.assertTrue(self.transaction._pending_orders)

    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=True,
    )
    def test_replace_order(self, mock__validate_controls):
        mock_order = mock.Mock()
        self.assertTrue(self.transaction.replace_order(mock_order, 1.01, 321))
        mock_order.replace.assert_called_with(1.01)
        mock__validate_controls.assert_called_with(mock_order, OrderPackageType.REPLACE)
        self.transaction._pending_replace = [(mock_order, None)]
        self.assertTrue(self.transaction._pending_orders)

    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=False,
    )
    def test_replace_order_violation(self, mock__validate_controls):
        mock_order = mock.Mock()
        self.assertFalse(self.transaction.replace_order(mock_order, 2.02))
        mock__validate_controls.assert_called_with(mock_order, OrderPackageType.REPLACE)
        self.transaction._pending_replace = []
        self.assertFalse(self.transaction._pending_orders)

    @mock.patch(
        "flumine.execution.transaction.Transaction._validate_controls",
        return_value=False,
    )
    def test_force_replace_order(self, mock__validate_controls):
        mock_order = mock.Mock()
        self.assertTrue(
            self.transaction.replace_order(mock_order, 1.01, 321, force=True)
        )
        mock_order.replace.assert_called_with(1.01)
        mock__validate_controls.assert_not_called()
        self.transaction._pending_replace = [(mock_order, None)]
        self.assertTrue(self.transaction._pending_orders)

    @mock.patch("flumine.execution.transaction.Transaction._create_order_package")
    def test_execute(self, mock__create_order_package):
        self.transaction._pending_orders = True
        mock_package = mock.Mock()
        mock__create_order_package.return_value = [mock_package]
        self.assertEqual(self.transaction.execute(), 0)
        mock_order = mock.Mock()
        self.transaction._pending_place = [(mock_order, 1234)]
        self.transaction._pending_cancel = [(mock_order, None)]
        self.transaction._pending_update = [(mock_order, None)]
        self.transaction._pending_replace = [(mock_order, 1234)]
        self.assertEqual(self.transaction.execute(), 4)
        mock__create_order_package.assert_has_calls(
            [
                call([(mock_order, 1234)], OrderPackageType.PLACE, async_=False),
                call([(mock_order, None)], OrderPackageType.CANCEL),
                call([(mock_order, None)], OrderPackageType.UPDATE),
                call([(mock_order, 1234)], OrderPackageType.REPLACE),
            ]
        )
        self.transaction.market.flumine.process_order_package.assert_has_calls(
            [
                call(mock__create_order_package()[0]),
                call(mock__create_order_package()[0]),
                call(mock__create_order_package()[0]),
                call(mock__create_order_package()[0]),
            ]
        )
        self.assertFalse(self.transaction._pending_orders)

    @mock.patch("flumine.execution.transaction.Transaction._create_order_package")
    def test_execute_async(self, mock__create_order_package):
        self.transaction._pending_orders = True
        self.transaction._async_place_orders = True
        mock_package = mock.Mock()
        mock__create_order_package.return_value = [mock_package]
        self.assertEqual(self.transaction.execute(), 0)
        mock_order = mock.Mock()
        self.transaction._pending_place = [(mock_order, 1234)]
        self.transaction._pending_cancel = [(mock_order, None)]
        self.transaction._pending_update = [(mock_order, None)]
        self.transaction._pending_replace = [(mock_order, 1234)]
        self.assertEqual(self.transaction.execute(), 4)
        mock__create_order_package.assert_has_calls(
            [
                call([(mock_order, 1234)], OrderPackageType.PLACE, async_=True),
                call([(mock_order, None)], OrderPackageType.CANCEL),
                call([(mock_order, None)], OrderPackageType.UPDATE),
                call([(mock_order, 1234)], OrderPackageType.REPLACE),
            ]
        )
        self.transaction.market.flumine.process_order_package.assert_has_calls(
            [
                call(mock__create_order_package()[0]),
                call(mock__create_order_package()[0]),
                call(mock__create_order_package()[0]),
                call(mock__create_order_package()[0]),
            ]
        )
        self.assertFalse(self.transaction._pending_orders)

    def test__validate_controls(self):
        mock_trading_control = mock.Mock()
        mock_client_control = mock.Mock()
        self.transaction.market.flumine.trading_controls = [mock_trading_control]
        self.transaction.market.flumine.client.trading_controls = [mock_client_control]
        mock_order = mock.Mock()
        mock_package_type = mock.Mock()
        self.assertTrue(
            self.transaction._validate_controls(mock_order, mock_package_type)
        )
        mock_trading_control.assert_called_with(mock_order, mock_package_type)
        mock_client_control.assert_called_with(mock_order, mock_package_type)

    def test__validate_controls_violation(self):
        mock_trading_control = mock.Mock()
        mock_trading_control.side_effect = ControlError("test")
        mock_client_control = mock.Mock()
        self.transaction.market.flumine.trading_controls = [mock_trading_control]
        self.transaction.market.flumine.client.trading_controls = [mock_client_control]
        mock_order = mock.Mock()
        mock_package_type = mock.Mock()
        self.assertFalse(
            self.transaction._validate_controls(mock_order, mock_package_type)
        )
        mock_trading_control.assert_called_with(mock_order, mock_package_type)
        mock_client_control.assert_not_called()

    @mock.patch("flumine.execution.transaction.BetfairOrderPackage")
    def test__create_order_package(self, mock_betfair_order_package):
        mock_betfair_order_package.order_limit.return_value = 2
        packages = self.transaction._create_order_package(
            [(1, None), (2, None), (3, 123), (4, 123), (5, 123)], OrderPackageType.PLACE
        )
        mock_betfair_order_package.assert_has_calls(
            [
                call(
                    client=self.transaction.market.flumine.client,
                    market_id=self.transaction.market.market_id,
                    orders=[1, 2],
                    package_type=OrderPackageType.PLACE,
                    bet_delay=self.transaction.market.market_book.bet_delay,
                    market_version=None,
                    async_=False,
                ),
                call(
                    client=self.transaction.market.flumine.client,
                    market_id=self.transaction.market.market_id,
                    orders=[3, 4],
                    package_type=OrderPackageType.PLACE,
                    bet_delay=self.transaction.market.market_book.bet_delay,
                    market_version=123,
                    async_=False,
                ),
                call(
                    client=self.transaction.market.flumine.client,
                    market_id=self.transaction.market.market_id,
                    orders=[5],
                    package_type=OrderPackageType.PLACE,
                    bet_delay=self.transaction.market.market_book.bet_delay,
                    market_version=123,
                    async_=False,
                ),
            ]
        )
        self.assertEqual(len(packages), 3)
        mock_betfair_order_package.order_limit.assert_called_with(
            OrderPackageType.PLACE
        )

    @mock.patch("flumine.execution.transaction.Transaction.execute")
    def test_enter_exit(self, mock_execute):
        with self.transaction as t:
            self.assertEqual(self.transaction, t)
            t._pending_orders = True
        mock_execute.assert_called()



import unittest
from unittest import mock

from flumine.backtest import utils


class NewDateTimeTest(unittest.TestCase):
    @mock.patch("flumine.backtest.utils.config")
    def test_new_date_time(self, mock_config):
        mock_config.current_time = 123
        x = utils.NewDateTime
        self.assertEqual(x.utcnow(), 123)


class SimulatedDateTimeTest(unittest.TestCase):
    def setUp(self):
        self.s = utils.SimulatedDateTime()

    def test_init(self):
        self.assertIsNone(self.s._real_datetime)

    @mock.patch("flumine.backtest.utils.config")
    def test_reset_real_datetime(self, mock_config):
        mock_real_datetime = mock.Mock()
        self.s._real_datetime = mock_real_datetime
        self.s.reset_real_datetime()
        mock_real_datetime.utcnow.assert_called()
        self.assertEqual(mock_config.current_time, mock_real_datetime.utcnow())

    @mock.patch("flumine.backtest.utils.config")
    def test_call(self, mock_config):
        mock_dt = mock.Mock()
        self.s(mock_dt)
        self.assertEqual(mock_config.current_time, mock_dt)

    @mock.patch("flumine.backtest.utils.config")
    def test_context_manager(self, mock_config):
        with self.s as datetime:
            mock_config.current_time = 456
            self.assertEqual(datetime.utcnow(), 456)
        import datetime

        self.assertIsInstance(datetime.datetime.utcnow(), datetime.datetime)



import re
import uuid
import json
import logging
import hashlib
import datetime
import functools
from pathlib import Path
from typing import Optional, Tuple, Callable, Union
from decimal import Decimal, ROUND_HALF_UP
from betfairlightweight.resources.bettingresources import MarketBook, RunnerBook

from . import config
from .exceptions import FlumineException

logger = logging.getLogger(__name__)

CUTOFFS = (
    (2, 100),
    (3, 50),
    (4, 20),
    (6, 10),
    (10, 5),
    (20, 2),
    (30, 1),
    (50, 0.5),
    (100, 0.2),
    (1000, 0.1),
)
MIN_PRICE = 1.01
MAX_PRICE = 1000
MARKET_ID_REGEX = re.compile(r"1.\d{9}")
EVENT_ID_REGEX = re.compile(r"\d{8}")
STRATEGY_NAME_HASH_LENGTH = 13


def detect_file_type(file_path: Union[str, tuple]) -> str:
    if isinstance(file_path, tuple):
        file_path = file_path[0]
    path_name = Path(file_path).name
    market_match = bool(MARKET_ID_REGEX.match(path_name))
    event_match = bool(EVENT_ID_REGEX.match(path_name))
    if market_match and not event_match:
        return "MARKET"
    elif not market_match and event_match:
        return "EVENT"
    else:
        return "UNKNOWN"


def create_short_uuid() -> str:
    return str(uuid.uuid4())[:8]


def file_line_count(file_path: str) -> int:
    with open(file_path) as f:
        for i, l in enumerate(f):
            pass
    return i + 1


def get_file_md(file_dir: Union[str, tuple], value: str) -> Optional[str]:
    # get value from raw streaming file marketDefinition
    if isinstance(file_dir, tuple):
        file_dir = file_dir[0]
    with open(file_dir, "r") as f:
        first_line = f.readline()
        update = json.loads(first_line)
    if "mc" not in update or not isinstance(update["mc"], list) or not update["mc"]:
        return None
    md = update["mc"][0].get("marketDefinition", {})
    return md.get(value)


def chunks(l: list, n: int) -> list:
    for i in range(0, len(l), n):
        yield l[i : i + n]


def create_cheap_hash(txt: str, length: int = 15) -> str:
    # This is just a hash for debugging purposes.
    #    It does not need to be unique, just fast and short.
    # https://stackoverflow.com/questions/14023350
    hash_ = hashlib.sha1()
    hash_.update(txt.encode())
    return hash_.hexdigest()[:length]


def as_dec(value):
    return Decimal(str(value))


def arange(start, stop, step):
    while start < stop:
        yield start
        start += step


def make_prices(min_price, cutoffs):
    prices = []
    cursor = as_dec(min_price)
    for cutoff, step in cutoffs:
        prices.extend(arange(as_dec(cursor), as_dec(cutoff), as_dec(1 / step)))
        cursor = cutoff
    prices.append(as_dec(MAX_PRICE))
    return prices


PRICES = make_prices(MIN_PRICE, CUTOFFS)
PRICES_FLOAT = [float(price) for price in PRICES]


def get_nearest_price(price, cutoffs=CUTOFFS):
    if price <= MIN_PRICE:
        return MIN_PRICE
    if price > MAX_PRICE:
        return MAX_PRICE
    price = as_dec(price)
    for cutoff, step in cutoffs:
        if price < cutoff:
            break
    step = as_dec(step)
    return float((price * step).quantize(2, ROUND_HALF_UP) / step)


def get_price(data: list, level: int) -> Optional[float]:
    try:
        return data[level]["price"]
    except KeyError:
        return
    except IndexError:
        return
    except TypeError:
        return


def get_size(data: list, level: int) -> Optional[float]:
    try:
        return data[level]["size"]
    except KeyError:
        return
    except IndexError:
        return
    except TypeError:
        return


def get_sp(runner: RunnerBook) -> Optional[float]:
    if isinstance(runner.sp, list):
        return
    elif runner.sp is None:
        return
    elif runner.sp.actual_sp == "NaN":
        return
    else:
        return runner.sp.actual_sp


@functools.lru_cache(maxsize=2048)
def price_ticks_away(price: float, n_ticks: int) -> float:
    try:
        price_index = PRICES_FLOAT.index(price)
        new_index = price_index + n_ticks
        if new_index < 0:
            return 1.01
        return PRICES_FLOAT[new_index]
    except IndexError:
        return 1000


def calculate_matched_exposure(mb: list, ml: list) -> Tuple:
    """Calculates exposure based on list
    of (price, size)
    returns the tuple (profit_if_win, profit_if_lose)
    """
    if not mb and not ml:
        return 0.0, 0.0
    back_exp, back_profit, lay_exp, lay_profit = 0, 0, 0, 0
    for p, s in mb:
        back_exp += -s
        back_profit += (p - 1) * s
    for p, s in ml:
        lay_exp += (p - 1) * -s
        lay_profit += s
    _win = back_profit + lay_exp
    _lose = lay_profit + back_exp
    return round(_win, 2), round(_lose, 2)


def calculate_unmatched_exposure(ub: list, ul: list) -> Tuple:
    """Calculates worse-case exposure based on list
    of (price, size)
    returns the tuple (profit_if_win, profit_if_lose)

    The worst case profit_if_win arises if all lay bets are matched and the selection wins.
    The worst case profit_if_lose arises if all back bets are matched and the selection loses.

    """
    if not ub and not ul:
        return 0.0, 0.0
    back_exp, lay_exp = 0, 0
    for p, s in ub:
        back_exp += -s
    for p, s in ul:
        lay_exp += (p - 1) * -s
    return round(lay_exp, 2), round(back_exp, 2)


def wap(matched: list) -> Tuple[float, float]:
    if not matched:
        return 0, 0
    a, b = 0, 0
    for _, p, s in matched:
        a += p * s
        b += s
    if b == 0 or a == 0:
        return 0, 0
    else:
        return round(b, 2), round(a / b, 2)


def call_strategy_error_handling(
    func: Callable, market, market_book: MarketBook
) -> Optional[bool]:
    try:
        return func(market, market_book)
    except FlumineException as e:
        logger.error(
            "FlumineException %s in %s (%s)" % (e, func.__name__, market.market_id),
            exc_info=True,
        )
    except Exception as e:
        logger.critical(
            "Unknown error %s in %s (%s)" % (e, func.__name__, market.market_id),
            exc_info=True,
        )
        if config.raise_errors:
            raise
    return False


def call_middleware_error_handling(middleware, market) -> None:
    try:
        middleware(market)
    except FlumineException as e:
        logger.error(
            "FlumineException %s in %s (%s)" % (e, middleware, market.market_id),
            exc_info=True,
        )
    except Exception as e:
        logger.critical(
            "Unknown error %s in %s (%s)" % (e, middleware, market.market_id),
            exc_info=True,
        )
        if config.raise_errors:
            raise


def call_process_orders_error_handling(strategy, market, strategy_orders: list) -> None:
    try:
        strategy.process_orders(market, strategy_orders)
    except FlumineException as e:
        logger.error(
            "FlumineException %s in %s (%s)" % (e, strategy, market.market_id),
            exc_info=True,
        )
    except Exception as e:
        logger.critical(
            "Unknown error %s in %s (%s)" % (e, strategy, market.market_id),
            exc_info=True,
        )
        if config.raise_errors:
            raise


def get_runner_book(
    market_book: MarketBook, selection_id: int, handicap=0
) -> Optional[RunnerBook]:
    """Returns runner book based on selection id."""
    for runner_book in market_book.runners:
        if (
            runner_book.selection_id == selection_id
            and runner_book.handicap == handicap
        ):
            return runner_book


def get_market_notes(market, selection_id: int) -> Optional[str]:
    """Returns a string of notes for a runner,
    currently 'back,lay,last_price_traded'
    """
    runner = get_runner_book(market.market_book, selection_id)
    if runner:
        return "{0},{1},{2}".format(
            get_price(runner.ex.available_to_back, 0),
            get_price(runner.ex.available_to_lay, 0),
            runner.last_price_traded,
        )


def get_event_ids(markets: list, event_type_id: str) -> list:
    event_ids = []
    for market in markets:
        if not market.closed and market.event_type_id == event_type_id:
            event_ids.append(market.event_id)
    return list(set(event_ids))


def create_time(publish_time: int, id_: str) -> datetime.datetime:
    pt_datetime = datetime.datetime.utcfromtimestamp(publish_time / 1e3)
    event_id, start_time = id_.split(".")
    hour, minute = int(start_time[:2]), int(start_time[2:])
    return pt_datetime.replace(hour=hour, minute=minute, second=0, microsecond=0)
